@@ -1,30 +1,35 @@
 import collections.abc
 
-from typing import Sequence, overload, Union, Iterable, Iterator, Any, Optional, Tuple, Callable
+from pathlib import Path
+from typing import Sequence, overload, Union, Iterable, Iterator, Any, Optional, Tuple, Callable, Mapping, Type
 from coba.backports import Literal
 
+from coba            import pipes
+from coba.contexts   import CobaContext, DiskCacher
 from coba.random     import CobaRandom
-from coba.pipes      import Pipes, Source, HttpSource, ListSource, JsonDecode
+from coba.pipes      import Pipes, Source, HttpSource, IterableSource, JsonDecode
 from coba.exceptions import CobaException
 
-from coba.environments.filters     import EnvironmentFilter
-from coba.environments.filters     import Binary, Shuffle, Take, Sparse, Reservoir, Cycle, Scale
-from coba.environments.filters     import Impute, Where, Noise, Riffle, Sort
+from coba.environments.filters   import EnvironmentFilter, Repr, Batch, Chunk
+from coba.environments.filters   import Binary, Shuffle, Take, Sparse, Reservoir, Cycle, Scale
+from coba.environments.filters   import Impute, Where, Noise, Riffle, Sort, Flatten, Cache, Params, Grounded
 from coba.environments.templates import EnvironmentsTemplateV1, EnvironmentsTemplateV2
 
-from coba.environments          .primitives import Environment
-from coba.environments.logged   .primitives import LoggedEnvironment
-from coba.environments.simulated.primitives import SimulatedEnvironment
-from coba.environments.warmstart.primitives import WarmStartEnvironment
+from coba.environments.primitives import Environment, Context, Action
 
 from coba.environments.simulated.synthetics import LinearSyntheticSimulation, NeighborsSyntheticSimulation
-from coba.environments.simulated.synthetics import KernelSyntheticSimulation, MLPSyntheticSimulation
+from coba.environments.simulated.synthetics import KernelSyntheticSimulation, MLPSyntheticSimulation, LambdaSimulation
 from coba.environments.simulated.openml     import OpenmlSimulation
 from coba.environments.simulated.supervised import SupervisedSimulation
 
-class Environments:
+class Environments (collections.abc.Sequence):
     """A friendly wrapper around commonly used environment functionality."""
 
+    @staticmethod
+    def cache_dir(path:Union[str,Path]) -> Type['Environments']:
+        CobaContext.cacher = DiskCacher(path)
+        return Environments
+
     @overload
     @staticmethod
     def from_template(filesource:Source[Iterable[str]]) -> 'Environments': ...
@@ -47,7 +52,6 @@ class Environments:
             except:
                 raise e
 
-
     @staticmethod
     def from_prebuilt(name:str) -> 'Environments':
         """Instantiate Environments from a pre-built definition made for diagnostics and comparisons across projects."""
@@ -67,7 +71,7 @@ class Environments:
         definition_txt = definition_txt.replace('"./', f'"{repo_url}/{name}/')
         definition_txt = definition_txt.replace('.json"', '.json?raw=True"')
 
-        return Environments.from_template(ListSource([definition_txt]))
+        return Environments.from_template(IterableSource([definition_txt]))
 
     @staticmethod
     def from_linear_synthetic(
@@ -147,7 +151,6 @@ class Environments:
     @overload
     @staticmethod
     def from_openml(data_id: Union[int,Sequence[int]],
-        cat_as_str: bool = False,
         drop_missing: bool = True,
         take: int = None) -> 'Environments':
         ...
@@ -155,7 +158,6 @@ class Environments:
     @overload
     @staticmethod
     def from_openml(*,task_id: Union[int,Sequence[int]],
-        cat_as_str: bool = False,
         drop_missing: bool = True,
         take: int = None) -> 'Environments':
         ...
@@ -164,7 +166,7 @@ class Environments:
     def from_openml(*args,**kwargs) -> 'Environments':
         """Create a SimulatedEnvironment from datasets available on openml."""
 
-        kwargs.update(zip(['data_id','cat_as_str','drop_missing','take'], args))
+        kwargs.update(zip(['data_id','drop_missing','take'], args))
 
         if 'data_id' in kwargs and isinstance(kwargs['data_id'],int):
             kwargs['data_id'] = [kwargs['data_id']]
@@ -201,6 +203,28 @@ class Environments:
         """Create a SimulatedEnvironment from a supervised dataset"""
         return Environments(SupervisedSimulation(*args, **kwargs))
 
+    @overload
+    def from_lambda(self,
+        n_interactions: Optional[int],
+        context       : Callable[[int               ],Context         ],
+        actions       : Callable[[int,Context       ],Sequence[Action]],
+        reward        : Callable[[int,Context,Action],float           ]) -> 'Environments':
+        """Create a SimulatedEnvironment from lambda functions."""
+
+    @overload
+    def from_lambda(self,
+        n_interactions: Optional[int],
+        context       : Callable[[int               ,CobaRandom],Context         ],
+        actions       : Callable[[int,Context       ,CobaRandom],Sequence[Action]],
+        reward        : Callable[[int,Context,Action,CobaRandom],float           ],
+        seed          : int) -> 'Environments':
+        """Create a SimulatedEnvironment from lambda functions."""
+
+    @staticmethod
+    def from_lambda(*args,**kwargs) -> 'Environments':
+        """Create a SimulatedEnvironment from lambda functions."""
+        return Environments(LambdaSimulation(*args,**kwargs))
+
     def __init__(self, *environments: Union[Environment, Sequence[Environment]]):
         """Instantiate an Environments class.
 
@@ -223,8 +247,28 @@ class Environments:
         """Convert an environment from a dense representation to sparse. This has little utility beyond debugging."""
         return self.filter(Sparse(context,action))
 
-    def shuffle(self, seeds: Union[Sequence[int],int] = 1) -> 'Environments':
+    @overload
+    def shuffle(self, seed: int = 1) -> 'Environments':
+        ...
+
+    @overload
+    def shuffle(self, seeds: Iterable[int]) -> 'Environments':
+        ...
+
+    @overload
+    def shuffle(self, *, n:int) -> 'Environments':
+        ...
+
+    def shuffle(self, *args,**kwargs) -> 'Environments':
         """Shuffle the order of the interactions in the Environments."""
+
+        flat = lambda a: next(pipes.Flatten().filter([a]))
+
+        if kwargs:
+            seeds = range(kwargs['n'])
+        else:
+            seeds = flat(args) or [1]
+                
         if isinstance(seeds,int): seeds = [seeds]
         return self.filter([Shuffle(seed) for seed in seeds])
 
@@ -240,18 +284,23 @@ class Environments:
         """Cycle all rewards associated with actions by one place."""
         return self.filter(Cycle(after))
 
+    def params(self, params: Mapping[str,Any]) -> 'Environments':
+        """Add params to the environments."""
+        return self.filter(Params(params))
+
     def take(self, n_interactions: int) -> 'Environments':
         """Take a fixed number of interactions from the Environments."""
         return self.filter(Take(n_interactions))
 
-    def reservoir(self, n_interactions: int, seeds: Sequence[int]) -> 'Environments':
+    def reservoir(self, n_interactions: int, seeds: Union[int,Sequence[int]]) -> 'Environments':
         """Take a random fixed number of interactions from the Environments."""
+        if isinstance(seeds,int): seeds = [seeds]
         return self.filter([Reservoir(n_interactions,seed) for seed in seeds])
 
     def scale(self,
-        shift: Union[float,Literal["min","mean","med"]] = 0,
+        shift: Union[float,Literal["min","mean","med"]] = "min",
         scale: Union[float,Literal["minmax","std","iqr","maxabs"]] = "minmax",
-        target: Literal["features","rewards"] = "features",
+        target: Literal["context","rewards","argmax"] = "context",
         using: Optional[int] = None) -> 'Environments':
         """Apply an affine shift and scaling factor to precondition environments."""
         return self.filter(Scale(shift, scale, target, using))
@@ -271,26 +320,60 @@ class Environments:
         action : Callable[[float,CobaRandom], float] = None,
         reward : Callable[[float,CobaRandom], float] = None,
         seed   : int = 1) -> 'Environments':
-        """Add noise to an environments context, actions and rewards."""
-        return self.filter(Noise(context,action,reward,seed))
+        """Add noise to an environment's context, actions and rewards."""
+        return Environments(*self).filter(Noise(context,action,reward,seed))
+
+    def flat(self) -> 'Environments':
+        """Flatten environment's context and actions."""
+        return self.filter(Flatten())
+
+    def materialize(self) -> 'Environments':
+        """Materialize the environments in memory. Ideal for stateful environments such as Jupyter Notebook."""
+        #we use pies.cache directly because the environment cache will copy 
+        #which we don't need when we materialize an environment in memory
+        envs = Environments([Pipes.join(env, pipes.Cache(25)) for env in self])
+        for env in envs: list(env.read()) #force read to pre-load cache
+        return envs
+
+    def grounded(self, n_users: int, n_normal:int, n_words:int, n_good:int, seed:int=1) -> 'Environments':
+        """Convert from simulated environments to interaction grounded environments."""
+        return self.filter(Grounded(n_users, n_normal, n_words, n_good, seed))
+
+    def repr(self, 
+        cat_context:Literal["onehot","onehot_tuple","string"] = "onehot",
+        cat_actions:Literal["onehot","onehot_tuple","string"] = "onehot") -> 'Environments':
+        """Determine how certain types of data is represented."""
+        return self.filter(Repr(cat_context,cat_actions))
+
+    def batch(self, batch_size:int) -> 'Environments':
+        """Batch interactions for learning and evaluation."""
+        return self.filter(Batch(batch_size))
+
+    def chunk(self, cache:bool = True) -> 'Environments':
+        """Create a chunk point in the environments pipeline to compose how multiprocess is conducted."""
+        envs = Environments([Pipes.join(env, Chunk()) for env in self])
+        return envs.cache() if cache else envs
+
+    def cache(self) -> 'Environments':
+        """Create a cache point in the environments so that earlier steps in the pipeline can be re-used in several pipes."""
+        return Environments([Pipes.join(env, Cache(25)) for env in self])
 
     def filter(self, filter: Union[EnvironmentFilter,Sequence[EnvironmentFilter]]) -> 'Environments':
         """Apply filters to each environment currently in Environments."""
         filters = filter if isinstance(filter, collections.abc.Sequence) else [filter]
-        self._environments = [ Pipes.join(e,f) for e in self._environments for f in filters ]
-        return self
+        return Environments([Pipes.join(e,f) for e in self._environments for f in filters])
 
-    def __getitem__(self, index:int) -> Union[SimulatedEnvironment, LoggedEnvironment, WarmStartEnvironment]:
+    def __getitem__(self, index:int) -> Environment:
         return self._environments[index]
 
-    def __iter__(self) -> Iterator[Union[SimulatedEnvironment, LoggedEnvironment, WarmStartEnvironment]]:
-        return self._environments.__iter__()
+    def __iter__(self) -> Iterator[Environment]:
+        return iter(self._environments)
 
     def __len__(self) -> int:
         return len(self._environments)
 
     def __add__(self, other: 'Environments') -> 'Environments':
-        return Environments(self._environments,other._environments)
+        return Environments(self._environments+other._environments)
 
     def __str__(self) -> str:
         return "\n".join([f"{i+1}. {e}" for i,e in enumerate(self._environments)])
