@@ -8,33 +8,34 @@ from numbers import Number
 from operator import truediv, sub, gt
 from abc import abstractmethod
 from itertools import chain, repeat, accumulate, groupby, count
-from typing import Any, Dict, List, Set, Tuple, Optional, Sequence, Hashable, Iterator, Union, Type, Callable, NamedTuple
+from typing import Any, Dict, List, Set, Tuple, Optional, Sequence, Hashable, Iterable, Iterator, Union, Type, Callable, NamedTuple
 from coba.backports import Literal
 
 from coba.statistics import Mean, StandardDeviation, StandardErrorOfMean, BootstrapConfidenceInterval, BinomialConfidenceInterval, PointAndInterval
 from coba.contexts import CobaContext
 from coba.exceptions import CobaException
 from coba.utilities import PackageChecker
-from coba.pipes import Pipes, Sink, Source, JsonEncode, JsonDecode, DiskSource, DiskSink, ListSource, ListSink, Foreach
+from coba.pipes import Pipes, Sink, Source, JsonEncode, JsonDecode, DiskSource, DiskSink, IterableSource, ListSink, Foreach
 
-def exponential_moving_average(values:Sequence[float], span:int=None):
+def exponential_moving_average(values:Sequence[float], span:int=None) -> Iterable[float]:
     #exponential moving average identical to Pandas df.ewm(span=span).mean()
     alpha = 2/(1+span)
     cumwindow  = list(accumulate(values          , lambda a,v: v + (1-alpha)*a))
     cumdivisor = list(accumulate([1.]*len(values), lambda a,v: v + (1-alpha)*a)) #type: ignore
-    return list(map(truediv, cumwindow, cumdivisor))
+    return map(truediv, cumwindow, cumdivisor)
 
-def moving_average(values:Sequence[float], span:int=None):
+def moving_average(values:Sequence[float], span:int=None) -> Iterable[float]:
 
     if span == 1: 
         return values
 
     if span is None or span >= len(values):
-        return [a/n for n,a in enumerate(accumulate(values),1)]
+        return (a/n for n,a in enumerate(accumulate(values),1))
 
     window_sums  = accumulate(map(sub, values, chain(repeat(0,span),values)))
     window_sizes = chain(range(1,span), repeat(span))
-    return list(map(truediv,window_sums,window_sizes))
+
+    return map(truediv,window_sums,window_sizes)
 
 class Table:
     """A container class for storing tabular data."""
@@ -272,8 +273,8 @@ class TransactionIO_V3(Source['Result'], Sink[Any]):
 
         self._log_file = log_file
         self._minify   = minify
-        self._source   = DiskSource(log_file) if log_file else ListSource()
-        self._sink     = DiskSink(log_file) if log_file else ListSink(self._source.items)
+        self._source   = DiskSource(log_file) if log_file else IterableSource()
+        self._sink     = DiskSink(log_file) if log_file else ListSink(self._source.iterable)
 
     def write(self, item: Any) -> None:
         if isinstance(self._sink, ListSink):
@@ -289,7 +290,7 @@ class TransactionIO_V3(Source['Result'], Sink[Any]):
         sim_rows = {}
         int_rows = {}
 
-        if isinstance(self._source, ListSource):
+        if isinstance(self._source, IterableSource):
             decoded_source = self._source
         else:
             decoded_source = Pipes.join(self._source, Foreach(JsonDecode()))
@@ -342,8 +343,8 @@ class TransactionIO_V4(Source['Result'], Sink[Any]):
 
         self._log_file = log_file
         self._minify   = minify
-        self._source   = DiskSource(log_file) if log_file else ListSource()
-        self._sink     = DiskSink(log_file)   if log_file else ListSink(self._source.items)
+        self._source   = DiskSource(log_file) if log_file else IterableSource()
+        self._sink     = DiskSink(log_file)   if log_file else ListSink(self._source.iterable)
 
     def write(self, item: Any) -> None:
         if isinstance(self._sink, ListSink):
@@ -359,7 +360,7 @@ class TransactionIO_V4(Source['Result'], Sink[Any]):
         env_rows = {}
         int_rows = {}
 
-        if isinstance(self._source, ListSource):
+        if isinstance(self._source, IterableSource):
             decoded_source = self._source
         else:
             decoded_source = Pipes.join(self._source, Foreach(JsonDecode()))
@@ -454,7 +455,7 @@ class Plotter:
         ylim: Optional[Tuple[Number,Number]],
         xticks: bool,
         yticks: bool,
-        filename: Optional[str]) -> None:
+        out: Union[None,Literal['screen'],str]) -> None:
         pass
 
 class MatplotlibPlotter(Plotter):
@@ -471,7 +472,7 @@ class MatplotlibPlotter(Plotter):
         yticks: bool,
         xrotation: Optional[float],
         yrotation: Optional[float],
-        filename: Optional[str]
+        out: Union[None,Literal['screen'],str]
     ) -> None:
 
         PackageChecker.matplotlib('Result.plot_learners')
@@ -480,7 +481,6 @@ class MatplotlibPlotter(Plotter):
         color_cycle = plt.rcParams['axes.prop_cycle'].by_key()['color']
 
         ax: plt.Axes
-        show = ax is None
 
         bad_xlim  = xlim and xlim[0] is not None and xlim[1] is not None and xlim[0] >= xlim[1]
         bad_ylim  = ylim and ylim[0] is not None and ylim[1] is not None and ylim[0] >= ylim[1]
@@ -492,9 +492,15 @@ class MatplotlibPlotter(Plotter):
             if bad_ylim:
                 CobaContext.logger.log("The ylim end is less than the ylim start. Plotting is impossible.")
             if bad_lines:
-                CobaContext.logger.log(f"No data was found for plotting in the given results.")
+                CobaContext.logger.log(f"No data was found for plotting.")
         else:
-            ax = ax or plt.figure(num="coba").add_subplot(111) #type: ignore
+
+            if not ax or isinstance(ax,int):
+                if 'coba' in plt.get_figlabels():
+                    f = plt.figure(num="coba")
+                    ax = f.add_subplot(111) if not isinstance(ax,int) else f.add_subplot(ax)
+                else:
+                    ax = plt.subplot(111) if not isinstance(ax,int) else plt.subplot(ax)
 
             in_lim = lambda v,lim: lim is None or ((lim[0] or -float('inf')) <= v and v <= (lim[1] or float('inf')))
 
@@ -534,7 +540,7 @@ class MatplotlibPlotter(Plotter):
 
             padding = .05
             ax.margins(0)
-            ax.set_xticks([min(ax.get_xlim()[1], max(ax.get_xlim()[0],x)) for x in ax.get_xticks()])
+            ax.set_xticks(ax.get_xticks() if not xlim else [min(ax.get_xlim()[1], max(ax.get_xlim()[0],x)) for x in ax.get_xticks()])
             ax.margins(padding)
 
             if xlim:
@@ -565,13 +571,14 @@ class MatplotlibPlotter(Plotter):
             if not yticks:
                 plt.yticks([])
 
-            if show:
+            if out not in ['screen',None]:
+                plt.tight_layout()
+                plt.savefig(str(out), dpi=300)
                 plt.show()
                 plt.close()
 
-            if filename:
-                plt.tight_layout()
-                plt.savefig(filename, dpi=300)
+            if out=="screen":
+                plt.show()
                 plt.close()
 
 class FilterPlottingData:
@@ -605,10 +612,8 @@ class FilterPlottingData:
 class SmoothPlottingData:
 
     def filter(self, rows:Sequence[Dict[str,Any]], y:str, span:Optional[int]) -> Sequence[Dict[str,Any]]:
-
         e_key = "environment_id"
         l_key = "learner_id"
-
         return [{e_key:row[e_key], l_key:row[l_key], y:moving_average(row[y],span)} for row in rows]
 
 class ContrastPlottingData:
@@ -654,11 +659,15 @@ class TransformToXYE:
             avg = Mean()
             YE = lambda z: (avg.calculate(z) if len(z) > 1 else z[0], None)
 
-        is_scatter = isinstance(rows[0][y][0],(list,tuple))
+        iters = [ iter(row[y]) for row in rows ]
+        first_val = next(iters[0])
+        iters[0] = chain([first_val],iters[0])
+
+        is_scatter = isinstance(first_val,(list,tuple))
 
         if x == ['index']:
-            Z = list(zip(*[row[y] for row in rows]))
-            X = list(range(1,len(Z)+1))
+            Z = zip(*iters)
+            X = count(1)
 
             if not is_scatter:
                 points = [(x,None)+YE(z) for x,z in zip(X,Z) ]
@@ -667,8 +676,8 @@ class TransformToXYE:
         else:
             XZ = collections.defaultdict(list)
             make_x = lambda env: env[x[0]] if len(x) == 1 else tuple(env[k] for k in x)
-            for row in rows:
-                XZ[str(make_x(envs[row["environment_id"]]))].append(row[y][-1])
+            for row,I in zip(rows,iters):
+                XZ[str(make_x(envs[row["environment_id"]]))].append(list(I)[-1])
 
             if not is_scatter:
                 points = [(x,None)+YE(z) for x,z in XZ.items()]
@@ -833,8 +842,31 @@ class Result:
         ylim       : Tuple[Number,Number] = None,
         xticks     : bool = True,
         yticks     : bool = True,
-        filename   : str = None,
+        out        : Union[None,Literal['screen'],str] = 'screen',
         ax = None) -> None:
+        """Plot a direct contrast of the performance for two learners.
+
+        Args:
+            learner_id1: The first learner to plot in the contrast.
+            learner_id2: The second learner to plot in the contrast.
+            x: The value to plot on the x-axis. This can either be index or environment columns to group by.
+            y: The value to plot on the y-axis.
+            mode: The kind of contrast plot to make: diff plots the pairwise difference, prob plots the the probability
+                of learner_id1 beating learner_id2, and scatter plots learner_id1 on x-axis and learner_id2 on y axis.
+            span: The number of y values to smooth together when reporting y. If this is None then the average of all y
+                values up to current is shown otherwise a moving average with window size of span (the window will be
+                smaller than span initially).
+            err: This determines what kind of error bars to plot (if any). If `None` then no bars are plotted, if 'se'
+                the standard error is shown, and if 'sd' the standard deviation is shown.
+            labels: The legend labels to use in the plot. These should be in order of the actual legend labels.
+            colors: The colors used to plot the learners plot.
+            xlim: Define the x-axis limits to plot. If `None` the x-axis limits will be inferred.
+            ylim: Define the y-axis limits to plot. If `None` the y-axis limits will be inferred.
+            xticks: Whether the x-axis labels should be drawn.
+            yticks: Whether the y-axis labels should be drawn.
+            out: Indicate where the plot should be sent to after plotting is finished.
+            ax: Provide an optional axes that the plot will be drawn to. If not provided a new figure/axes is created.
+        """
 
         x = [x] if isinstance(x,str) else list(x)
         self._validate_parameters(x)
@@ -865,9 +897,10 @@ class Result:
         fmt = "-" if x == ['index'] else "."
 
         plots = []
-        if loss: plots.append(Points(*zip(*loss), 2, 1, labels[1] + " " + f"({len(loss)})", fmt))
-        if tie : plots.append(Points(*zip(*tie) , 1, 1, 'Tie'     + " " + f"({len(tie )})", fmt))
-        if win : plots.append(Points(*zip(*win) , 0, 1, labels[0] + " " + f"({len(win )})", fmt))        
+
+        if loss: plots.append(Points(*zip(*loss), colors[0], 1, labels[1] + " " + f"({len(loss)})", fmt))
+        if tie : plots.append(Points(*zip(*tie) , colors[1], 1, 'Tie'     + " " + f"({len(tie )})", fmt))
+        if win : plots.append(Points(*zip(*win) , colors[2], 1, labels[0] + " " + f"({len(win )})", fmt))
 
         if mode != 'scat':
             leftmost_x  = (loss+tie+win)[ 0][0]
@@ -881,7 +914,7 @@ class Result:
         yrotation = 0
 
         if mode != "scat":
-            xlabel = "Interactions" if x==['index'] else x[0] if len(x) == 1 else x
+            xlabel = "Interaction" if x==['index'] else x[0] if len(x) == 1 else x
             ylabel = f"{labels[0]} - {labels[1]}" if mode=="diff" else f"P({labels[0]} > {labels[1]})"
         else:
             xlabel = y
@@ -889,26 +922,28 @@ class Result:
 
         title  = f"{ylabel} ({len(rows) if x==['index'] else len(XYE)} Environments)"
 
-        self._plotter.plot(ax, plots, title, xlabel, ylabel, xlim, ylim, xticks, yticks, xrotation, yrotation, filename)
+        self._plotter.plot(ax, plots, title, xlabel, ylabel, xlim, ylim, xticks, yticks, xrotation, yrotation, out)
 
     def plot_learners(self,
-        x       : Union[str,Sequence[str]] = "index",
-        y       : str = "reward",
-        span    : int = None,
-        err     : Union[Literal['se','sd','bs'], None, PointAndInterval] = None,
-        labels  : Sequence[str] = None,
-        colors  : Sequence[str] = None,
-        xlim    : Tuple[Number,Number] = None,
-        ylim    : Tuple[Number,Number] = None,
-        xticks  : bool = True,
-        yticks  : bool = True,
-        filename: str = None,
+        ids   : Union[int,Sequence[int]] = None, 
+        x     : Union[str,Sequence[str]] = "index",
+        y     : str = "reward",
+        span  : int = None,
+        err   : Union[Literal['se','sd','bs'], None, PointAndInterval] = None,
+        labels: Sequence[str] = None,
+        colors: Union[int,Sequence[Union[str,int]]] = None,
+        xlim  : Tuple[Number,Number] = None,
+        ylim  : Tuple[Number,Number] = None,
+        xticks: bool = True,
+        yticks: bool = True,
+        out   : Union[None,Literal['screen'],str] = 'screen',
         ax = None) -> None:
         """Plot the performance of multiple learners on multiple environments. It gives a sense of the expected
             performance for different learners across independent environments. This plot is valuable in gaining
             insight into how various learners perform in comparison to one another.
 
         Args:
+            ids: Sequence of learner ids to plot (if None we will plot all learnesr in the result).
             x: The value to plot on the x-axis. This can either be index or environment columns to group by.
             y: The value to plot on the y-axis.
             span: The number of y values to smooth together when reporting y. If this is None then the average of all y
@@ -916,21 +951,29 @@ class Result:
                 smaller than span initially).
             err: This determines what kind of error bars to plot (if any). If `None` then no bars are plotted, if 'se'
                 the standard error is shown, and if 'sd' the standard deviation is shown.
+            labels: The legend labels to use in the plot. These should be in order of the actual legend labels.
+            colors: The colors used to plot the learners plot.
             xlim: Define the x-axis limits to plot. If `None` the x-axis limits will be inferred.
             ylim: Define the y-axis limits to plot. If `None` the y-axis limits will be inferred.
-            labels: The legend labels to use in the plot. These should be in order of the actual legend labels.
-            filename: Provide a filename to write plot image to disk.
+            xticks: Whether the x-axis labels should be drawn.
+            yticks: Whether the y-axis labels should be drawn.
+            out: Indicate where the plot should be sent to after plotting is finished.
             ax: Provide an optional axes that the plot will be drawn to. If not provided a new figure/axes is created.
         """
 
-        x = [x] if isinstance(x,str) else list(x)
+        if isinstance(ids,int): ids = [ids]
+        if isinstance(colors,int): colors = [colors]
+        if isinstance(labels,str): labels = [labels]
+        if isinstance(x,str): x = [x]
+
         self._validate_parameters(x)
 
-        rows = FilterPlottingData().filter(list(self.interactions), x, y)
+        interactions = self.interactions if not ids else self.filter_lrn(learner_id=ids).interactions
+
+        rows = FilterPlottingData().filter(list(interactions), x, y)
         rows = SmoothPlottingData().filter(rows, y, span)
 
         env_rows = self.environments
-        n_envs   = len(set([row['environment_id'] for row in rows]))
         get_key  = lambda row: row['learner_id']
         lines    = []
 
@@ -939,12 +982,12 @@ class Result:
         for i, (lrn_id, lrn_rows) in enumerate(groupby(sorted(rows, key=get_key),key=get_key)):
             lrn_rows = list(lrn_rows)
             XYE      = TransformToXYE().filter(lrn_rows, env_rows, x, y, err)
-            color    = colors[i] if colors and i < len(colors) else i
-            label    = labels[i] if labels and i < len(labels) else self.learners[lrn_id]['full_name']
+            color    = i if not colors else i+max(colors) if isinstance(colors[0],int) else colors[i] if i < len(colors) else i
+            label    = labels[i] if labels and i < len(labels) else self.learners[lrn_id]['full_name']            
             lines.append(Points(*zip(*XYE), color, 1, label, style))
 
         lines  = sorted(lines, key=lambda line: -line[1][-1])
-        xlabel = "Interactions" if x==['index'] else x[0] if len(x) == 1 else x
+        xlabel = "Interaction" if x==['index'] else x[0] if len(x) == 1 else x
         ylabel = y.capitalize().replace("_pct"," Percent")
 
         title = ("Instantaneous" if span == 1 else f"Span {span}" if span else "Progressive") + f" {ylabel}"
@@ -953,7 +996,7 @@ class Result:
         if x != ['index']:
             title = "Final " + title
 
-        self._plotter.plot(ax, lines, title, xlabel, ylabel, xlim, ylim, xticks, yticks, 0, 0, filename)
+        self._plotter.plot(ax, lines, title, xlabel, ylabel, xlim, ylim, xticks, yticks, 0, 0, out)
 
     def __str__(self) -> str:
         return str({"Learners": len(self._learners), "Environments": len(self._environments), "Interactions": len(self._interactions) })
