@@ -6,9 +6,8 @@ from coba.backports import Literal
 from coba.exceptions import CobaException
 from coba.random import CobaRandom
 from coba.environments import Context, Action
-from coba.contexts import InteractionContext
 
-from coba.learners.primitives import Learner, SafeLearner, Probs, Info
+from coba.learners.primitives import Learner, SafeLearner, PMF, kwargs, Actions
 
 class CorralLearner(Learner):
     """A meta-learner that takes a collection of learners and determines
@@ -39,10 +38,10 @@ class CorralLearner(Learner):
                 learning rate to shrink towards 0 slowly. A value of inf means that the learning rate
                 will remain constant.
             mode: Determines the method with which feedback is provided to the base learners. The
-                original paper used importance sampling. We also support `off-policy` and `rejection`.
+                original paper used importance sampling. We also support `off-policy`.
             seed: A seed for a random number generation in ordre to get repeatable results.
         """
-        if mode not in ["importance", "off-policy", "rejection"]:
+        if mode not in ["importance", "off-policy"]:
             raise CobaException("The provided `mode` for CorralLearner was unrecognized.")
 
         self._base_learners = [ SafeLearner(learner) for learner in learners]
@@ -62,40 +61,29 @@ class CorralLearner(Learner):
         self._mode = mode
 
         self._random_pick   = CobaRandom(seed)
-        self._random_reject = CobaRandom(CobaRandom(seed).randint(0,10000))
+        self._random_reject = CobaRandom(CobaRandom(seed).randint(0,1000))
 
     @property
     def params(self) -> Dict[str, Any]:
         return { "family": "corral", "eta": self._eta_init, "mode":self._mode, "T": self._T, "B": [ str(b) for b in self._base_learners ], "seed":self._random_pick._seed }
 
-    def predict(self, context: Context, actions: Sequence[Action]) -> Tuple[Probs, Info]:
+    def predict(self, context: Context, actions: Sequence[Action]) -> Tuple[PMF,kwargs]:
 
         base_predicts = [ base_algorithm.predict(context, actions) for base_algorithm in self._base_learners ]
-        base_predicts, base_infos = zip(*base_predicts)
+        base_actions, base_probs, base_infos = zip(*base_predicts)
 
-        if self._mode in ["importance"]:
-            base_actions = [ self._random_pick.choice(actions, predict) for predict in base_predicts              ]
-            base_probs   = [ predict[actions.index(action)] for action,predict in zip(base_actions,base_predicts) ]
+        predict = [ sum([p_b*int(a==b_a) for p_b,b_a in zip(self._p_bars, base_actions)]) for a in range(len(actions)) ]
+        info    = (base_actions, base_probs, base_infos)
 
-            predict = [ sum([p_b*int(a==b_a) for p_b,b_a in zip(self._p_bars, base_actions)]) for a in actions ]
-            info    = (base_actions, base_probs, base_infos, base_predicts, actions, predict)
+        return (predict, {'info':info})
 
-        if self._mode in ["off-policy", "rejection"]:
-            predict = [ sum([p_b*b_p[i] for p_b,b_p in zip(self._p_bars, base_predicts)]) for i in range(len(actions)) ]
-            info    = (None, None, base_infos, base_predicts, actions, predict)
-
-        return (predict, info)
-
-    def learn(self, context: Context, action: Action, reward: float, probability:float, info: Info) -> None:
+    def learn(self, context: Context, actions: Actions, action: Action, reward: float, probability:float, **kwargs) -> None:
 
         assert  0 <= reward and reward <= 1, "This Corral implementation assumes a loss between 0 and 1"
 
-        base_actions = info[0]
-        base_probs   = info[1]
-        base_infos   = info[2]
-        base_preds   = info[3]
-        actions      = info[4]
-        predict      = info[5]
+        base_actions = kwargs['info'][0]
+        base_probs   = kwargs['info'][1]
+        base_infos   = kwargs['info'][2]
 
         if self._mode == "importance":
             # This is what is in the original paper. It has the following characteristics:
@@ -105,7 +93,7 @@ class CorralLearner(Learner):
             # The reward, R, supplied to the base learners satisifies E[R|context,A] = E[reward|context,A]
             for learner, A, P, base_info in zip(self._base_learners, base_actions, base_probs, base_infos):
                 R = reward * int(A==action)/probability
-                learner.learn(context, A, R, P, base_info)
+                learner.learn(context, actions, A, R, P, **base_info)
 
         if self._mode == "off-policy":
             # An alternative variation to the paper is provided below. It has the following characterisitcs:
@@ -113,32 +101,11 @@ class CorralLearner(Learner):
             #   > It uses a MVUB reward estimator (aka, the unmodified, observed reward)
             #   > It is "off-policy" (i.e., base learners receive action feedback distributed differently from their predicts).
             for learner, base_info in zip(self._base_learners, base_infos):
-                learner.learn(context, action, reward, probability, base_info)
-
-        if self._mode == "rejection":
-            # An alternative variation to the paper is provided below. It has the following characterisitcs:
-            #   > It doesn't necessarily provide feedback to every base learner on every iteration
-            #   > It uses a MVUB reward estimator (aka, the unmodified, observed reward) when it does provide feedback
-            #   > It is "on-policy" (i.e., base learners receive action feedback is distributed identically to their predicts).
-            p = self._random_reject.random() #can I reuse this across all learners like this??? I think so???
-            for learner, base_info, base_predict in zip(self._base_learners, base_infos, base_preds):
-                f = lambda a: base_predict[actions.index(a)] #the PMF we want
-                g = lambda a: predict[actions.index(a)]      #the PMF we have
-
-                M = max([f(A)/g(A) for A in actions if g(A) > 0])
-                if p <= f(action)/(M*g(action)):
-                    learner.learn(context, action, reward, f(action), base_info)
-
-        # Instant loss is an unbiased estimate of E[loss|learner] for this iteration.
-        # Our estimate differs from the orginal Corral paper because we have access to the
-        # action probabilities of the base learners while the Corral paper did not assume
-        # access to this information. This information allows for a loss esimator with the same
-        # expectation as the original Corral paper's estimator but with a lower variance.
+                learner.learn(context, actions, action, reward, probability, **base_info)
 
         loss = 1-reward
 
-        picked_index = actions.index(action)
-        instant_loss = [ loss * base_pred[picked_index]/probability for base_pred in base_preds ]
+        instant_loss = [ loss/probability * (base_action==action) for base_action in base_actions ]
         self._ps     = CorralLearner._log_barrier_omd(self._ps, instant_loss, self._etas)
         self._p_bars = [ (1-self._gamma)*p + self._gamma*1/len(self._base_learners) for p in self._ps ]
 
@@ -147,12 +114,6 @@ class CorralLearner(Learner):
                 self._rhos[i] = 2/self._p_bars[i]
                 self._etas[i] *= self._beta
 
-        base_predict_data = { f"predict_{i}": base_preds[i][picked_index] for i in range(len(self._base_learners)) }
-        base_pbar_data    = { f"pbar_{i}"   : self._p_bars[i]             for i in range(len(self._base_learners)) }
-        predict_data      = { "predict"     : probability, **base_predict_data, **base_pbar_data }
-
-        InteractionContext.learner_info.update({**predict_data, **base_predict_data, **base_pbar_data})
-
     @staticmethod
     def _log_barrier_omd(ps, losses, etas) -> Sequence[float]:
 
