@@ -1,7 +1,9 @@
 import warnings
 import importlib
 
+from itertools import chain
 from collections import defaultdict
+from typing import TypeVar, Iterable, Tuple, Sequence
 
 from coba.exceptions import CobaExit
 
@@ -102,15 +104,6 @@ class PackageChecker:
     def _handle_import_error(caller_name:str, pkg_name:str):
         coba_exit(f"ERROR: {caller_name} requires the {pkg_name} package. You can install this package via `pip install {pkg_name}`.")
 
-class HashableDict(dict):
-    def __init__(self, *args, **kwargs) -> None:
-        super().__init__(*args, **kwargs)
-        self._hash = hash(tuple(self.items()))
-
-    def __hash__(self) -> int:
-        assert self._hash == hash(tuple(self.items()))
-        return self._hash
-
 class KeyDefaultDict(defaultdict):
     def __missing__(self, key):
         if self.default_factory is None:
@@ -119,3 +112,12 @@ class KeyDefaultDict(defaultdict):
             value = self.default_factory(key)
             self[key] = value
             return value
+
+_T = TypeVar("_T")
+def peek_first(items: Iterable[_T]) -> Tuple[_T, Iterable[_T]]:
+    items = iter(items)
+    try:
+        first = next(items)
+        return first, chain([first],items)
+    except StopIteration:
+        return None, []
